from .api.client import Client
from .api.resources.admin_resources import AdminResources
from .api.resources.auth_resources import AuthResources
from .api.rehive_util import RehiveUtil
from .api.resources.user_resources import UserResources
from .api.resources.transaction_resource import APITransactions
from .api.resources.accounts_resources import APIAccounts
from .api.resources.company_resources import APICompany


class Rehive:

    def __init__(self, token=None, connection_pool_size=0, network="live", **kwargs):
        # API Classes
        # Leave token blank if logging in
        self.client = Client(token, connection_pool_size, network, **kwargs)
        self.admin = AdminResources(self.client)
        self.auth = AuthResources(self.client)
        self.util = RehiveUtil(self.client)
        self.user = UserResources(self.client)
        self.transactions = APITransactions(self.client)
        self.accounts = APIAccounts(self.client)
        self.company = APICompany(self.client)



from .base_resources import Resource, ResourceCollection, ResourceList


class APIAccounts(ResourceList, ResourceCollection):
    def __init__(self, client, endpoint='', filters=None, resource_identifier=None):
        self.resources = (APIAdminCurrencies,)
        super(APIAccounts, self).__init__(client, endpoint, filters)
        self.create_resources(self.resources)

    @classmethod
    def get_resource_name(cls):
        return 'accounts'


class APIAdminCurrencies(ResourceList):
    def __init__(self, client, endpoint, filters=None):
        super(APIAdminCurrencies, self).__init__(client, endpoint, filters)

    def make_active_currency(self, code):
        return self.patch(code, active=True)

    @classmethod
    def get_resource_name(cls):
        return 'currencies'



import re
import copy
from urllib.parse import urlparse
from urllib.parse import urljoin
import urllib.parse


class Resource(object):
    def __init__(self, client, endpoint, filters=None):
        self.client = client
        self.endpoint = endpoint + self.get_resource_name() + '/'
        self.filters = filters
        self.resource_identifier = ''
        self.has_been_hydrated = False

    def get(self, function=None, **kwargs):
        url = self._build_url(function, **kwargs)
        response = self.client.get(url)
        return self._handle_resource_data(response)

    def post(self, data={}, function=None, idempotent_key=None, **kwargs):
        # Allow us to parse through arbitrary request arguments
        request_kwargs = {}
        if 'file' in kwargs:
            request_kwargs['files'] = {
                'file': (open(kwargs.get('file'), 'rb'))
            }
            kwargs.pop('file', None)
        # We need this flag to force non-json on file uploads
        json = True
        if 'json' in kwargs:
            json = kwargs.get('json')
            kwargs.pop('json', None)
        data = {**data, **kwargs}
        url = self._build_url(function)
        response = self.client.post(
            url,
            data,
            json=json,
            idempotent_key=idempotent_key,
            **request_kwargs
        )
        return self._handle_resource_data(response)

    def put(self, function='', idempotent_key=None, **kwargs):
        data = kwargs
        url = self._build_url(function)
        response = self.client.put(
            url,
            data,
            idempotent_key=idempotent_key
        )
        return self._handle_resource_data(response)

    def patch(self, function='', idempotent_key=None, **kwargs):
        data = kwargs
        url = self._build_url(function)
        response = self.client.patch(
            url,
            data,
            idempotent_key=idempotent_key
        )
        return self._handle_resource_data(response)

    def delete(self, function='', **kwargs):
        data = kwargs
        url = self._build_url(function)
        response = self.client.delete(url, data)
        return self._handle_resource_data(response)

    def update(self, function='', idempotent_key=None, **kwargs):
        return self.patch(
            function,
            idempotent_key=idempotent_key,
            **kwargs
        )

    def create(self, idempotent_key=None, **kwargs):
        return self.post(
            idempotent_key=idempotent_key,
            **kwargs
        )

    # PRIVATE METHODS
    def _build_url(self, function=None, **kwargs):
        # currently pagination should override all
        if kwargs.get('pagination'):
            endpoint = urllib.parse.urljoin(self.endpoint, kwargs.get('pagination'))
        else:
            if function:
                endpoint = urllib.parse.urljoin(
                    self.endpoint, function
                )
            else:
                endpoint = self.endpoint
            endpoint = self._append_trailing_slash(endpoint)
            if (self.resource_identifier is not None):
                endpoint = urllib.parse.urljoin(
                    endpoint, self.resource_identifier
                )
            endpoint = self._append_trailing_slash(endpoint)
        if kwargs.get('filters'):
            filters = kwargs.get('filters')
            filters_string = urllib.parse.urlencode(filters)
            endpoint = endpoint + '?' + filters_string
        return endpoint

    def _put_or_patch_by_identifier(self,
                                    method,
                                    identifier,
                                    identifier_field,
                                    **kwargs):
        if (identifier is not None and identifier_field is not None):
            function = '?' + identifier_field + '=' + identifier
        else:
            raise Exception('No identifier supplied')
        data = kwargs
        url = self._build_url(function)
        if method == 'PATCH':
            response = self.client.patch(url, data)
        elif method == 'PUT':
            response = self.client.put(url, data)
        else:
            raise Exception('No method supplied')
        return response

    def _append_trailing_slash(self, url):
        if (re.search(r'\/+$', url) is None):
            url = url + '/'
        return url

    def _set_resource_identifier(self, resource_identifiter):
        self.resource_identifier = resource_identifiter

    def _set_endpoint(self, endpoint):
        self.endpoint = self._append_trailing_slash(endpoint)

    def get_resource_name(cls):
        raise NotImplementedError(
            'The resource should define its own string name'
        )

    def _handle_resource_data(self, response):
        return_data = response
        if 'data' in response:
            return_data = response.get('data')
        self.has_been_hydrated = True

        return return_data


class ResourceList(Resource):

    def __init__(self, client, endpoint, filters=None):
        super(ResourceList, self).__init__(client, endpoint, filters)
        self.next = None
        self.previous = None
        self._count = 0

    def get(self, endpoint=None, **kwargs):
        url = self._build_url(endpoint, **kwargs)
        response = self.client.get(url)
        return self._handle_pagination_data(response)

    def get_next(self, **kwargs):
        url = self._build_url(pagination=self.next, **kwargs)
        response = self.client.get(url)
        return self._handle_pagination_data(response)

    def get_previous(self, **kwargs):
        url = self._build_url(pagination=self.previous, **kwargs)
        response = self.client.get(url)
        return self._handle_pagination_data(response)

    @property
    def count(self):
        if (self.has_been_hydrated is False):
            self.get()
        return self._count

    # PRIVATE METHODS
    def _handle_pagination_data(self, response):
        data = response.get('data')
        return_data = data
        if 'next' in data:
            if data.get('next') is not None:
                self.next = self._get_next_page_filter(data.get('next'))
            else:
                self.next = data.get('next')
        if 'previous' in data:
            if data.get('previous') is not None:
                self.previous = self._get_next_page_filter(data['previous'])
            else:
                self.previous = data.get('previous')
        if 'count' in data:
            self._count = data.get('count')
        if 'results' in data:
            return_data = data.get('results')
        self.has_been_hydrated = True

        return return_data

    def _get_next_page_filter(self, string):
        url_segments = string.split('/')
        last_segment = url_segments[-1]
        return last_segment

    def _build_pagination_url(self, pagination, function=None):
        endpoint = self.endpoint + function if function else self.endpoint
        paginatated_endpoint = endpoint + pagination
        if self.filters:
            paginatated_endpoint = paginatated_endpoint + self.filters
        return paginatated_endpoint


class ResourceCollection(object):

    def __init__(self):
        self.create_resources(self.resources)

    def create_resources(self, resources):
        for resource in resources:
            setattr(self,
                    self._sanitize_resource_name(
                        resource.get_resource_name()),
                    resource(self.client, self.endpoint))

    def obj(self, resource_identifiter):
        return self.object(resource_identifiter)

    def object(self, resource_identifiter):
        resource_object = copy.copy(self)
        resource_object._set_resource_identifier(resource_identifiter)
        resource_object._set_endpoint(resource_object.endpoint + resource_object.resource_identifier)
        resource_object.create_resources(resource_object.resources)
        return resource_object

    def _sanitize_resource_name(self, string):
        return string.replace('-', '_')



from .base_resources import Resource, ResourceCollection, ResourceList


class APICompany(Resource, ResourceCollection):
    def __init__(self, client, endpoint='', filters=None):
        self.resources = (
            APIBanks,
            APICurrencies,
            APIBankAccount,
        )
        super(APICompany, self).__init__(client, endpoint, filters)
        self.create_resources(self.resources)

    @classmethod
    def get_resource_name(cls):
        return 'company'


class APIBanks(ResourceList):

    @classmethod
    def get_resource_name(cls):
        return 'bank'


class APICurrencies(ResourceList):

    @classmethod
    def get_resource_name(cls):
        return 'currencies'


class APIBankAccount(Resource):

    @classmethod
    def get_resource_name(cls):
        return 'bank-account'



""" Python api for Rehive """
import os
import requests
import json

from .exception import APIException

API_ENDPOINT = os.environ.get("REHIVE_API_URL",
                              "https://rehive.com/api/3/")
API_STAGING_ENDPOINT = os.environ.get("API_STAGING_ENDPOINT",
                                      "https://staging.rehive.com/api/3/")


class Client:
    """
    Interface for interacting with the rehive api
    """

    def __init__(self,
                 token=None,
                 connection_pool_size=0,
                 network='live',
                 api_endpoint_url=None):

        self.token = token
        if api_endpoint_url:
            # Override the defaults
            self.endpoint = api_endpoint_url
        else:  
            self.endpoint = API_ENDPOINT if (network == 'live') else API_STAGING_ENDPOINT 
        self._connection_pool_size = connection_pool_size
        self._session = None

    def post(self, path, data, json=True, **kwargs):
        return self._request('post', path, data, json=json, **kwargs)

    def get(self, path, **kwargs):
        return self._request('get', path, **kwargs)

    def put(self, path, data, **kwargs):
        return self._request('put', path, data, **kwargs)

    def patch(self, path, data, **kwargs):
        return self._request('patch', path, data, **kwargs)

    def delete(self, path, data, **kwargs):
        return self._request('delete', path, **kwargs)

    def _create_session(self):
        self._session = requests.Session()
        if self._connection_pool_size > 0:
            adapter = requests.adapters.HTTPAdapter(
                pool_connections=self._connection_pool_size,
                pool_maxsize=self._connection_pool_size
            )
            self._session.mount('http://', adapter)
            self._session.mount('https://', adapter)

    def _request(self,
                 method,
                 path,
                 data=None,
                 json=True,
                 headers={},
                 idempotent_key=None,
                 **kwargs):
        if self._session is None:
            self._create_session()

        url = self.endpoint + path
        headers = self._get_headers(
            headers=headers,
            json=json,
            idempotent_key=idempotent_key
        )

        try:
            if (data and json):
                result = self._session.request(method,
                                               url,
                                               headers=headers,
                                               json=data,
                                               **kwargs)
            elif (data and not json):
                result = self._session.request(method,
                                               url,
                                               headers=headers,
                                               data=data,
                                               **kwargs)
            else:
                result = self._session.request(method, url, headers=headers, **kwargs)

            if not result.ok:
                if result.status_code == 404:
                    raise APIException('Not found: ' + url, result.status_code)
                error_data = result.json()
                raise APIException(error_data.get(
                        'message', 'General error'),
                        result.status_code, error_data)

            response_json = self._handle_result(result)
            return response_json

        except requests.exceptions.ConnectionError as e:
            raise APIException(str(e))
        except requests.exceptions.Timeout as e:
            raise APIException("Connection timed out.", None, str(e))
        except requests.exceptions.RequestException as e:
            raise APIException("General request error",  None, str(e))

    def _handle_result(self, result):
        json = result.json()

        # Check for token in response and set it for the current object
        if (json and 'data' in json and 'token' in json['data']):
            self.token = json['data']['token']

        return json

    def _get_headers(self, headers, json=True, idempotent_key=None):
        if json:
            headers['Content-Type'] = 'application/json'
        if self.token is not None:
            headers['Authorization'] = 'Token ' + str(self.token)
        if idempotent_key is not None:
            headers['Idempotency-Key'] = idempotent_key

        return headers



from .base_resources import ResourceList, Resource, ResourceCollection


class AdminResources(ResourceCollection):

    def __init__(self, client):
        # Api Client instance
        self.client = client
        self.endpoint = 'admin/'

        # Resources
        self.resources = (
            APIAdminUsers,
            APIAdminCurrencies,
            APIAdminTransactions,
            APIAdminAccounts,
            APIAdminCompany,
            APIAdminWebhooks,
            APIAdminSubtypes,
            APIAdminBankAccounts,
            APIAdminSwitches,
            APIAdminTiers,
            APIAdminPermissionGroups,
        )
        self.create_resources(self.resources)


class APIAdminNotifications(ResourceList):
    @classmethod
    def get_resource_name(cls):
        return 'notifications'


class APIAdminAccounts(ResourceList, ResourceCollection):
    def __init__(self, client, endpoint, filters=None, resource_identifier=None):
        self.resources = (APIAdminCurrencies,)
        super(APIAdminAccounts, self).__init__(client, endpoint, filters)

    @classmethod
    def get_resource_name(cls):
        return 'accounts'


class APIAdminCurrencies(ResourceList, ResourceCollection):
    def __init__(self, client, endpoint, filters=None):
        self.resources = (
            APIAdminLimits,
            APIAdminFees,
            APIGeneralSwitches,
            APIAdminBankAccounts,
            APIAdminOverview,
        )
        super(APIAdminCurrencies, self).__init__(client, endpoint, filters)

    def create(self, code, description, symbol, unit, divisibility):
        data = {
            "code": code,
            "description": description,
            "symbol": symbol,
            "unit": unit,
            "divisibility": divisibility
        }
        response = self.post(data)
        return response

    @classmethod
    def get_resource_name(cls):
        return 'currencies'


class APIAdminUsers(ResourceList, ResourceCollection):
    def __init__(self, client, endpoint, filters=None):
        self.resources = {
            APIAdminEmails,
            APIAdminMobiles,
            APIAdminCryptoAccounts,
            APIAdminSwitches,
            APIAdminDocuments,
            APIAdminAddresses,
            APIAdminOverview,
            APIAdminPermissions
        }
        super(APIAdminUsers, self).__init__(client, endpoint, filters)
        self.create_resources(self.resources)

    def create(self, first_name, last_name, email, mobile_number):
        data = {
            "first_name": first_name,
            "last_name": last_name,
            "email": email,
            "mobile_number": mobile_number
        }
        response = self.post(data)
        return response

    @classmethod
    def get_resource_name(cls):
        return 'users'


class APIAdminDocuments(ResourceList):
    def __init__(self, client, endpoint, filters=None):
        super(APIAdminDocuments, self).__init__(client, endpoint, filters)

    def create(self, document_type, file, **kwargs):
        return super().create(
            document_type=document_type,
            file=file,
            json=False,
            **kwargs
        )

    @classmethod
    def get_resource_name(cls):
        return 'documents'


class APIAdminEmails(ResourceList):
    def __init__(self, client, endpoint, filters=None):
        super(APIAdminEmails, self).__init__(client, endpoint, filters)

    def create(self, user, email, **kwargs):
        data = {
            'user': user,
            'email': email
        }
        return self.post(data, **kwargs)

    @classmethod
    def get_resource_name(cls):
        return 'emails'


class APIAdminAddresses(ResourceList):
    def __init__(self, client, endpoint, filters=None):
        super(APIAdminAddresses, self).__init__(client, endpoint, filters)

    @classmethod
    def get_resource_name(cls):
        return 'addresses'


class APIAdminOverview(ResourceList):
    def __init__(self, client, endpoint, filters=None):
        super(APIAdminOverview, self).__init__(client, endpoint, filters)

    @classmethod
    def get_resource_name(cls):
        return 'overview'


class APIAdminMobiles(ResourceList):
    def __init__(self, client, endpoint, filters=None):
        super(APIAdminMobiles, self).__init__(client, endpoint, filters)

    def create(self, user, number, **kwargs):
        data = {
            'user': user,
            'email': email
        }
        return self.post(data, **kwargs)

    @classmethod
    def get_resource_name(cls):
        return 'mobiles'


class APIAdminTransactions(ResourceList, ResourceCollection):
    def __init__(self, client, endpoint, filters=None):
        self.resources = {
            APIAdminSwitches,
            APIAdminWebhooks
        }
        super(APIAdminTransactions, self).__init__(client, endpoint, filters)
        self.create_resources(self.resources)

    def get_totals(self):
        response = self.get('totals/')
        return response

    def patch(self, tx_code, status, **kwargs):
        data = kwargs
        data['status'] = status
        return super(APIAdminTransactions, self).patch(tx_code + '/', **data)

    def confirm(self, tx_code, **kwargs):
        self.patch(tx_code, "complete", **kwargs)

    def fail(self, tx_code, **kwargs):
        self.patch(tx_code, "failed", **kwargs)

    def delete(self, tx_code, **kwargs):
        self.patch(tx_code, "deleted", **kwargs)

    def create_credit(self, user, amount, **kwargs):
        data = {
            'user': user,
            'amount': amount
        }
        response = self.post(data, 'credit/', **kwargs)
        return response

    def create_debit(self, user, amount, **kwargs):
        data = {
            'user': user,
            'amount': amount
        }
        response = self.post(data, 'debit/', **kwargs)
        return response

    def create_transfer(self, user, amount, recipient, **kwargs):
        data = {
            'user': user,
            'amount': amount,
            'recipient': recipient
        }
        response = self.post(data, 'transfer/', **kwargs)
        return response

    @classmethod
    def get_resource_name(cls):
        return 'transactions'


class APIAdminCompany(Resource, ResourceCollection):
    def __init__(self, client, endpoint, filters=None):
        self.resources = (APIGeneralSwitches,)
        super(APIAdminCompany, self).__init__(client, endpoint, filters)
        self.create_resources(self.resources)

    @classmethod
    def get_resource_name(cls):
        return 'company'


class APIAdminSwitches(ResourceList):
    def __init__(self, client, endpoint, filters=None):
        super(APIAdminSwitches, self).__init__(client, endpoint, filters)

    def create(self, switch_type, enabled=False, **kwargs):
        data = {
            'switch_type': switch_type,
            'enabled': enabled
        }
        return self.post(data, **kwargs)

    @classmethod
    def get_resource_name(cls):
        return 'switches'


class APIGeneralSwitches(APIAdminSwitches):
    def __init__(self, client, endpoint, filters=None):
        super(APIGeneralSwitches, self).__init__(client, endpoint, filters)


class APIAdminWebhooks(ResourceList):
    def __init__(self, client, endpoint, filters=None):
        super(APIAdminWebhooks, self).__init__(client, endpoint, filters)

    def create(self, url, **kwargs):
        data = {
            'url': url
        }
        return self.post(data, **kwargs)

    @classmethod
    def get_resource_name(cls):
        return 'webhooks'


class APIAdminTransactionWebhooks(ResourceList):
    def __init__(self, client, endpoint, filters=None):
        super(APIAdminTransactionWebhooks, self).__init__(client, endpoint, filters)

    def create(self, tx_type, url, **kwargs):
        data = {
            'tx_type': tx_type,
            'url': url
        }
        return self.post(data, **kwargs)

    @classmethod
    def get_resource_name(cls):
        return 'transactions'


class APIAdminSubtypes(ResourceList):
    def __init__(self, client, endpoint, filters=None):
        super(APIAdminSubtypes, self).__init__(client, endpoint, filters)

    def create(self, name, tx_type, **kwargs):
        data = {
            'name': name,
            'tx_type': tx_type
        }
        return self.post(data, **kwargs)

    @classmethod
    def get_resource_name(cls):
        return 'subtypes'


class APIAdminBankAccounts(ResourceList):
    def __init__(self, client, endpoint, filters=None):
        super(APIAdminBankAccounts, self).__init__(client, endpoint, filters)

    @classmethod
    def get_resource_name(cls):
        return 'bank-accounts'


class APIAdminLimits(ResourceList):
    def __init__(self, client, endpoint, filters=None):
        super(APIAdminLimits, self).__init__(client, endpoint, filters)

    def create(self, value, limit_type, tx_type, **kwargs):
        data = {
            'value': value,
            'type': limit_type,
            'tx_type': tx_type
        }
        return self.post(data, **kwargs)

    @classmethod
    def get_resource_name(cls):
        return 'limits'


class APIAdminFees(ResourceList):
    def __init__(self, client, endpoint, filters=None):
        super(APIAdminFees, self).__init__(client, endpoint, filters)

    def create(self, tx_type, **kwargs):
        data = {
            'tx_type': tx_type
        }
        return self.post(data, **kwargs)

    @classmethod
    def get_resource_name(cls):
        return 'fees'


class APIAdminTiers(ResourceList, ResourceCollection):
    def __init__(self, client, endpoint, filters=None):
        self.resources = (
            APIGeneralSwitches,
            APIAdminFees,
            APIAdminRequirements,
        )
        super(APIAdminTiers, self).__init__(client, endpoint, filters)

    def create(self, currency, **kwargs):
        data = {
            'currency': currency
        }
        return self.post(data, **kwargs)

    @classmethod
    def get_resource_name(cls):
        return 'tiers'


class APIAdminRequirements(ResourceList):
    def __init__(self, client, endpoint, filters=None):
        super(APIAdminRequirements, self).__init__(client, endpoint, filters)

    def create(self, requirement, **kwargs):
        data = {
            'requirement': requirement
        }
        return self.post(data, **kwargs)

    @classmethod
    def get_resource_name(cls):
        return 'requirements'


class APIAdminCryptoAccounts(ResourceList):
    def __init__(self, client, endpoint, filters=None):
        super(APIAdminCryptoAccounts, self).__init__(client, endpoint, filters)

    def create(self, address, crypto_type, **kwargs):
        return super().create(
            address=address,
            type=crypto_type,
            **kwargs
        )

    @classmethod
    def get_resource_name(cls):
        return 'crypto-accounts'


class APIAdminPermissions(ResourceList):
    def __init__(self, client, endpoint, filters=None):
        super(APIAdminPermissions, self).__init__(client, endpoint, filters)

    def create(self, type, level, **kwargs):
        return super().create(
            type=type,
            level=level,
            **kwargs
        )

    @classmethod
    def get_resource_name(cls):
        return 'permissions'


class APIAdminPermissionGroups(ResourceList, ResourceCollection):
    def __init__(self, client, endpoint, filters=None):
        self.resources = (
            APIAdminPermissions,
        )
        super(APIAdminPermissionGroups, self).__init__(client, endpoint, filters)

    def create(self, name, **kwargs):
        return super().create(
            name=name,
            **kwargs
        )

    @classmethod
    def get_resource_name(cls):
        return 'permission-groups'
